/****************************************************************
*                                                               *
*    Name: MyFtp.h                                              *
*                                                               *
*    Creator: Balazs Tuska                                      *
*                                                               *
*    Description:                                               *
*        This is the header file of the FTP commands.           *
*                                                               *
****************************************************************/
#ifndef __MYFTP_H_
#define __MYFTP_H_

#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <sys/time.h>
#include <time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include "Progress.h"
/* FtpAccess() type codes */
#define FTPLIB_DIR 1
#define FTPLIB_DIR_VERBOSE 2
#define FTPLIB_FILE_READ 3
#define FTPLIB_FILE_WRITE 4

/* FtpAccess() mode codes */
#define FTPLIB_ASCII 'A'
#define FTPLIB_IMAGE 'I'
#define FTPLIB_TEXT FTPLIB_ASCII
#define FTPLIB_BINARY FTPLIB_IMAGE

/* connection modes */
#define FTPLIB_PASSIVE 1
#define FTPLIB_PORT 2

#include <time.h>

/* connection option names */
#define FTPLIB_CONNMODE 1
#define FTPLIB_CALLBACK 2
#define FTPLIB_IDLETIME 3
#define FTPLIB_CALLBACKARG 4
#define FTPLIB_CALLBACKBYTES 5

#define FTP_XFER_IND        32768


#define FTPLIB_HOST_UNIX  1
#define FTPLIB_HOST_VMS   2
#define FTPLIB_HOST_NT    3

typedef struct NetBuf netbuf;
typedef int (*FtpCallback)(netbuf *nControl, int xfered, void *arg);

/* Moved from ftplib.c by Leo */
struct NetBuf {
    char *cput,*cget;
    int handle;
    int cavail,cleft;
    char *buf;
    int dir;
    netbuf *ctrl;
    int cmode;
    struct timeval idletime;
    FtpCallback idlecb;
    void *idlearg;
    int xfered;
    int cbbytes;
    int xfered1;
    char response[256];
    int host_type;
};
/****end moved****/

// -------------------------------------------------------------
// My Class
// -------------------------------------------------------------
class CMyFtp
{
public:
    CMyFtp();
    int Connect(const char *host, netbuf **nControl);
    int Options(int opt, long val, netbuf *nControl);
    int Login(const char *user, const char *pass, netbuf *nControl);
    int Mkdir(const char *path, netbuf *nControl);
    int Read(void *buf, int max, netbuf *nData);
    int Write(void *buf, int len, netbuf *nData);
    int Close(netbuf *nData);
    int Rmdir(const char *path, netbuf *nControl);
    int Chdir(const char *path, netbuf *nControl);
    int CDUp(netbuf *nControl);
    int Pwd(char *path, int max, netbuf *nControl);
    int Noop(netbuf *nControl);
    int Dir(const char *outputfile, const char *path, netbuf *nControl);
    int Size(const char *path, __int64 *size, char mode, netbuf *nControl);
    int ModDate(const char *path, char *dt, int max, netbuf *nControl);
    int Get(const char *output, const char *path, char mode,
        netbuf *nControl, __int64 size=-1,TProgressForm *PProgressForm=NULL);
    int Rename(const char *src, const char *dst, netbuf *nControl);
    int Delete(const char *fnm, netbuf *nControl);
    int Ls(char *output, char *path, netbuf *nControl);
    int Put(const char *input, const char *path, char mode,
        netbuf *nControl, __int64 fsz, __int64 size=-1,TProgressForm *PProgressForm=NULL);
    void Quit(netbuf *nControl);
    int SendCmd(const char *cmd, char expresp, netbuf *nControl);
    int Access(const char *path, int typ, int mode, netbuf *nControl,
        netbuf **nData);
    int FtpDelDir(netbuf* netb,char *from,char *TMPFILE,char *tmp1,char *ftpname);
    int socket_wait(netbuf *ctl);

private:
    int ftplib_debug;
    FtpCallback default_cb;

    void *memccpy(void *dest, const void *src, int c, int n);
    void ftp_error(const char* err);
    int readline(char *buf,int max,netbuf *ctl);
    int writeline(char *buf, int len, netbuf *nData);
    int readresp(char c, netbuf *nControl);
    void ftp_log_handler(const char* msg);
    int OpenPort(netbuf *nControl, netbuf **nData, int mode, int dir);
    int AcceptConnection(netbuf *nData, netbuf *nControl);
    int Xfer(const char *localfile, const char *path,
        netbuf *nControl, int typ, int mode, __int64 size=-1,TProgressForm *PProgressForm=NULL);
    int Ftp_convert_attr(char *sa);
    char *Ftp_getword(char *from,char* to);
    char *Ftp_nameword(char *from,char* to);
    int Ftp_getmonth(char *s);
    void Ftp_gettime(char *s,struct tm *ttm);
    char *strlwr (char *string);
    int FTP_what_month (char *month);
    int Ftp_calcwords(char* buf);

};

#endif

