/****************************************************************
*                                                               *
*    Name: Viewer.cpp                                           *
*                                                               *
*    Creator: Balazs Tuska                                      *
*                                                               *
*    Description:                                               *
*        This is the cpp file of the viewer class. This         *
*        class handle the files to view them.                   *
*                                                               *
****************************************************************/
//---------------------------------------------------------------------------

#include <clx.h>
#pragma hdrstop

#include "Viewer.h"
#include "Math.hpp"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.xfm"
TViewForm *ViewForm;
//---------------------------------------------------------------------------
__fastcall TViewForm::TViewForm(TComponent* Owner)
    : TForm(Owner)
{
    filesize=0;
    filename="";
    lwidth=500;
    lheight=50;
    line_num=0;

    herestream=0;
    hereline=0;
    herelinememo=0;

    findstring="";
    findfromhere=0;
    free=true;
}
//---------------------------------------------------------------------------
void __fastcall TViewForm::FormKeyDown(TObject *Sender, WORD &Key,
      TShiftState Shift)
{
// close: enter, esc, F3, F10, 'q'
    if( Key==4100 || Key==4101 || Key==4096 || Key==4146 || Key==4153 || Key==81 ) {
        Close();
    }

// F7,'F' - search
    if( Key==4150 || Key==70 ) {
        if( file ) FindInFile_my();
    }

// 'N' - next fit
    if( Key==78 ) {
        if( file ) FindInFile("*"+findstring+"*");
    }

    if( Key==4112 || Key==4113 || Key==4115 || Key==4117 || Key==4118 || Key==4119 ) {
        switch( Key ) {
// home
            case 4112:
                herestream=0;
                hereline=0;
                herelinememo=lheight;
                findfromhere=herestream;
                read(herestream);
                break;
// end
            case 4113:
                herestream=filesize-1;
                hereline=filesize;
                herelinememo=lheight;
                findfromhere=herestream;
                read(herestream);
                break;
// up
            case 4115:
                up();
                findfromhere=herestream;
                break;
// down
            case 4117:
                down();
                findfromhere=herestream;
                break;
// page up
            case 4118:
                up(lheight-2);
                findfromhere=herestream;
                break;
// page down
            case 4119:
                down(lheight-2);
                findfromhere=herestream;
                break;
        }
        Key=0;
        refresh();
        return;
    }

// wordwrap
    if( Key==4145 ) {
        Viewer_my->WordWrap=!Viewer_my->WordWrap;
    }
}
//---------------------------------------------------------------------------
void __fastcall TViewForm::FormShow(TObject *Sender)
{
    free=false;
    Font->Height=13;
    Viewer_my->Clear();
    Viewer_my->Font=Form1->viewerfont->Font;
    Viewer_my->Font->Height=Form1->viewerfontheight;
    Height=Form1->viewerheight;
    Width=Form1->viewerwidth;
    int readlimit=0;
    file=fopen64((filename).c_str(),"r");
    try {
        struct stat64 tf;
        if( lstat64((filename).c_str(),&tf)!=0 ) {}
        filesize=tf.st_size;
    } catch(...) {filesize=0;}
    ScrollBar1->Max=filesize;
    Caption="View - "+filename;
    lheight=Viewer_my->Height/(Viewer_my->Font->Height+1);
    readlimit=lheight*3;
    herestream=0;
    hereline=0;
    herelinememo=lheight;
    read(herestream);
    Viewer_my->Clear();
    refresh();
//    SetFocus();
    Viewer_my->SetFocus();
}
//---------------------------------------------------------------------------

void __fastcall TViewForm::Viewer_myKeyDown(TObject *Sender, WORD &Key,
      TShiftState Shift)
{
    FormKeyDown(Sender,Key,Shift);
}
//---------------------------------------------------------------------------

int TViewForm::read(int pos)
{
    if( !file ) {
        return 0;
    }
    char ch;
    AnsiString line="";
    int readed=0;

    if( viewer_lines!=NULL ) {
        delete []viewer_lines;
        viewer_lines=NULL;
    }
    viewer_lines=new AnsiString[lheight*3];
    int firstenter=pos;
    line_num=lheight;
    if( pos<0 ) pos=0;
    fseek(file,pos+1,SEEK_SET);
// forward reading
    while( line_num>=0 && ftell(file)>1 ) {
        fseek(file,ftell(file)-2,SEEK_SET);
        ch=fgetc(file);
        if( ch=='\n' || line.Length()>lwidth || ftell(file)==1) {
            if( ftell(file)==1 ) {
                line=String(ch)+line;
                readed++;
            }
            if( firstenter==pos )
                if( ftell(file)==1 && ch!='\n' )
                    firstenter=ftell(file)-1;
                else
                    firstenter=ftell(file);
            viewer_lines[line_num]=line;
            line_num--;
            line="";
        } else {
            line=String(ch)+line;
            readed++;
        }
    }
// backward reading
    line_num=lheight;
    herestream=firstenter;
    fseek(file,firstenter,SEEK_SET);
    while( line_num<lheight*3 && !feof(file) ) {
        ch=fgetc(file);
        if( ch=='\n' || line.Length()>lwidth || feof(file) ) {
            viewer_lines[line_num]=line;
            line_num++;
            line="";
        } else {
            line=line+ch;
            readed++;
        }
    }
}

void TViewForm::refresh()
{
    if( !file ) {
        return;
    }
    int percent=0;
    ScrollBar1->Position=herestream;
    if( filesize!=0 ) percent=int((float(herestream)/float(filesize))*100);
    status->SimpleText="Position/Size: "+IntToStr(herestream)+"/"+IntToStr(filesize)+"   "+IntToStr(percent)+"%";
    Viewer_my->Lines->BeginUpdate();
    Viewer_my->Lines->Clear();
    for( int i=herelinememo;i<herelinememo+lheight-2;i++ )
        Viewer_my->Lines->Add(viewer_lines[i]);
    Viewer_my->Lines->EndUpdate();
}

void TViewForm::up(int n)
{
    if( !file ) {
        return;
    }
    for( int i=0;i<n;i++ ) {
        if( herestream>0 ) {
            hereline--;
            herelinememo--;
            if( herelinememo<0 ) {
                herelinememo=lheight;
                read(herestream);
            }
            if( herestream-(viewer_lines[herelinememo].Length()+1)>=0 ) herestream-=(viewer_lines[herelinememo].Length()+1);
            else herestream=0;
        }
    }
}

void TViewForm::down(int n)
{
    if( !file ) {
        return;
    }
    for( int i=0;i<n;i++ ) {
        if( herestream+viewer_lines[herelinememo].Length()+1<filesize ) {
            hereline++;
            herestream+=viewer_lines[herelinememo].Length()+1;
            herelinememo++;
        }
        if( herelinememo>2*lheight ) {
            herelinememo=lheight;
            read(herestream+1);
        }
    }
}
void __fastcall TViewForm::ScrollBar1KeyDown(TObject *Sender, WORD &Key,
      TShiftState Shift)
{
    FormKeyDown(Sender,Key,Shift);
    Key=0;
    return;
}
//---------------------------------------------------------------------------

void __fastcall TViewForm::FormResize(TObject *Sender)
{
    ScrollBar1->Width=16;
    Viewer_my->Width=this->Width-1-ScrollBar1->Width;
    ScrollBar1->Left=this->Width-1-ScrollBar1->Width;
    Viewer_my->Height=this->Height-1-status->Height;
    ScrollBar1->Height=this->Height-2-status->Height;
    lheight=Viewer_my->Height/(Viewer_my->Font->Height+1);
    herelinememo=lheight;
    read(herestream);
    refresh();
}
//---------------------------------------------------------------------------

void TViewForm::FindInFile(AnsiString mask)
{
    if( !file ) {
        return;
    }
    __int64 lastst=0;
    __int64 lastst2=0;
    int mn=1;
    char ch;
    __int64 starthere=findfromhere;
    bool found=false;
    __int64 entercounter=0;
    __int64 entercounter_last=0;
    int inlinepos=findfromhere-herestream;
    if( inlinepos ) inlinepos++;
    int lastilp=0;
    int matchlen=0;
    int lastmatchlen=0;

    herestream=findfromhere;
    fseek(file,herestream,SEEK_SET);

    while( !feof(file) && herestream<=filesize && mn<=mask.Length() ) {
        fseek(file,herestream,SEEK_SET);
        ch=fgetc(file);
        if( mask[mn]=='*' ) {
            lastst=mn;
            lastst2=herestream;
            lastilp=inlinepos;
            lastmatchlen=matchlen;
            entercounter_last=entercounter;
            if( mn<mask.Length() ) {
                mn++;
                continue;
            } else {
                found=true;
                break;
            }
        }

        if( ch=='\n' ) {
            entercounter++;
            ch=' ';
            inlinepos=0;
            lastilp=0;
            matchlen=0;
            lastmatchlen=0;
        }
        if( Form1->ignorecase ) {
            if( ch>='A' && ch<='Z' )
                ch+=32;
            if( mask[mn]>='A' && mask[mn]<='Z' )
                mask[mn]+=32;
        }
        if( ch==mask[mn] || mask[mn]=='?' ) {
            mn++;
            matchlen++;
            if( mn>mask.Length() )
                if( feof(file) ) {
                    found=true;
                    break;
                } else
                    if( lastst==0 ) {
                        found=false;
                        break;
                    } else {
                        mn=lastst;
                        herestream=lastst2;
                        inlinepos=lastilp;
                        if( lastst==1 )
                            matchlen=lastmatchlen;
                        entercounter=entercounter_last;
                    }
        } else {
            if( lastst==0 ) {
                found=false;
                break;
            } else {
                mn=lastst;
                herestream=lastst2;
                inlinepos=lastilp;
                if( lastst==1 )
                    matchlen=lastmatchlen;
                else
                    if( lastmatchlen!=0 )
                        matchlen++;
                entercounter=entercounter_last;
            }
        }
        herestream++;
        inlinepos++;
    }
    if( found ) {
        hereline+=entercounter;
        herelinememo=lheight;
        findfromhere=herestream+mask.Length();
        read(herestream);
        refresh();
        Viewer_my->SelStart=inlinepos-matchlen-1;
        Viewer_my->SelLength=matchlen;
    } else {
        herestream=starthere;
        fseek(file,herestream,SEEK_SET);
    }
}

void TViewForm::FindInFile_my()
{
    if( !file ) {
        return;
    }
    actionform=&TActionForm(Owner);
    actionform->Caption="Find string(mask) in file";
    actionform->doon->Text=findstring;
    actionform->mess1->Caption="string(mask)";
    actionform->mess2->Caption="";
    actionform->status=0;
    actionform->symlinks->Enabled=false;
    actionform->doon2->Enabled=false;
    actionform->recursive->Enabled=false;
    actionform->presattr->Enabled=false;
    actionform->ShowModal();
// selection canceled
    if( actionform->status==2 ) {
        delete actionform;
        return;
    }
    findstring=String(actionform->doon->Text);
    findfromhere=herestream;
    FindInFile("*"+findstring+"*");
    delete actionform;
}

void __fastcall TViewForm::ScrollBar1Scroll(TObject *Sender,
      TScrollCode ScrollCode, int &ScrollPos)
{
    if( !file ) {
        return;
    }

    if( herestream+1==ScrollPos )
        down();
    else if( herestream-1==ScrollPos )
        up();
    else if( herestream+2==ScrollPos )
        down(lheight-2);
    else if( herestream-2==ScrollPos )
        up(lheight-2);
    else {
        herestream=ScrollBar1->Position;
        herelinememo=lheight;
        read(herestream);
    }
    refresh();
}
//---------------------------------------------------------------------------

void __fastcall TViewForm::FormClose(TObject *Sender, TCloseAction &Action)
{
    if( file ) {
        fclose(file);
    }
    if( viewer_lines!=NULL ) {
        delete []viewer_lines;
        viewer_lines=NULL;
    }
    filesize=0;
    filename="";
    free=true;
    Action=caFree;
}
//---------------------------------------------------------------------------

void __fastcall TViewForm::FormCreate(TObject *Sender)
{
    SetFocus();
}
//---------------------------------------------------------------------------

